"""
Provides a class called DLE to convert and solve dynamic linear economies
(as set out in Hansen & Sargent (2013)) as LQ problems.
"""

import numpy as np
from ._lqcontrol import LQ
from ._matrix_eqn import solve_discrete_lyapunov
from ._rank_nullspace import nullspace

class DLE(object):
    r"""
    This class is for analyzing dynamic linear economies, as set out in Hansen & Sargent (2013).
    The planner's problem is to choose \{c_t, s_t, i_t, h_t, k_t, g_t\}_{t=0}^\infty to maximize

        \max -(1/2) \mathbb{E} \sum_{t=0}^{\infty} \beta^t [(s_t - b_t).(s_t-b_t) + g_t.g_t]

    subject to the linear constraints

        \Phi_c c_t + \Phi_g g_t + \Phi_i i_t = \Gamma k_{t-1} + d_t
        k_t = \Delta_k k_{t-1} + \Theta_k i_t
        h_t = \Delta_h h_{t-1} + \Theta_h c_t
        s_t = \Lambda h_{t-1} + \Pi c_t

    and

        z_{t+1} = A_{22} z_t + C_2 w_{t+1}
        b_t = U_b z_t
        d_t = U_d z_t

    where h_{-1}, k_{-1}, and z_0 are given as initial conditions.

    Section 5.5 of HS2013 describes how to map these matrices into those of
    a LQ problem.

    HS2013 sort the matrices defining the problem into three groups:

    Information: A_{22}, C_2, U_b , and U_d characterize the motion of information
    sets and of taste and technology shocks

    Technology: \Phi_c, \Phi_g, \Phi_i, \Gamma, \Delta_k, and \Theta_k determine the
    technology for producing consumption goods

    Preferences: \Delta_h, \Theta_h, \Lambda, and \Pi determine the technology for
    producing consumption services from consumer goods. A scalar discount factor \beta
    determines the preference ordering over consumption services.

    Parameters
    ----------
    Information : tuple
        Information is a tuple containing the matrices A_{22}, C_2, U_b, and U_d
    Technology : tuple
        Technology is a tuple containing the matrices \Phi_c, \Phi_g, \Phi_i, \Gamma,
        \Delta_k, and \Theta_k
    Preferences : tuple
        Preferences is a tuple containing the scalar \beta and the
        matrices \Lambda, \Pi, \Delta_h, and \Theta_h

    """

    def __init__(self, information, technology, preferences):

        # === Unpack the tuples which define information, technology and preferences === #
        self.a22, self.c2, self.ub, self.ud = information
        self.phic, self.phig, self.phii, self.gamma, self.deltak, self.thetak = technology
        self.beta, self.llambda, self.pih, self.deltah, self.thetah = preferences

        # === Computation of the dimension of the structural parameter matrices === #
        self.nb, self.nh = self.llambda.shape
        self.nd, self.nc = self.phic.shape
        self.nz, self.nw = self.c2.shape
        _, self.ng = self.phig.shape
        self.nk, self.ni = self.thetak.shape

        # === Creation of various useful matrices === #
        uc = np.hstack((np.eye(self.nc), np.zeros((self.nc, self.ng))))
        ug = np.hstack((np.zeros((self.ng, self.nc)), np.eye(self.ng)))
        phiin = np.linalg.inv(np.hstack((self.phic, self.phig)))
        phiinc = uc.dot(phiin)
        b11 = - self.thetah.dot(phiinc).dot(self.phii)
        a1 = self.thetah.dot(phiinc).dot(self.gamma)
        a12 = np.vstack((self.thetah.dot(phiinc).dot(
            self.ud), np.zeros((self.nk, self.nz))))

        # === Creation of the A Matrix for the state transition of the LQ problem === #

        a11 = np.vstack((np.hstack((self.deltah, a1)), np.hstack(
            (np.zeros((self.nk, self.nh)), self.deltak))))
        self.A = np.vstack((np.hstack((a11, a12)), np.hstack(
            (np.zeros((self.nz, self.nk + self.nh)), self.a22))))

        # === Creation of the B Matrix for the state transition of the LQ problem === #

        b1 = np.vstack((b11, self.thetak))
        self.B = np.vstack((b1, np.zeros((self.nz, self.ni))))

        # === Creation of the C Matrix for the state transition of the LQ problem === #

        self.C = np.vstack((np.zeros((self.nk + self.nh, self.nw)), self.c2))

        # === Define R,W and Q for the payoff function of the LQ problem === #

        self.H = np.hstack((self.llambda, self.pih.dot(uc).dot(phiin).dot(self.gamma), self.pih.dot(
            uc).dot(phiin).dot(self.ud) - self.ub, -self.pih.dot(uc).dot(phiin).dot(self.phii)))
        self.G = ug.dot(phiin).dot(
            np.hstack((np.zeros((self.nd, self.nh)), self.gamma, self.ud, -self.phii)))
        self.S = (self.G.T.dot(self.G) + self.H.T.dot(self.H)) / 2

        self.nx = self.nh + self.nk + self.nz
        self.n = self.ni + self.nh + self.nk + self.nz

        self.R = self.S[0:self.nx, 0:self.nx]
        self.W = self.S[self.nx:self.n, 0:self.nx]
        self.Q = self.S[self.nx:self.n, self.nx:self.n]

        # === Use quantecon's LQ code to solve our LQ problem === #

        lq = LQ(self.Q, self.R, self.A, self.B,
                self.C, N=self.W, beta=self.beta)

        self.P, self.F, self.d = lq.stationary_values()

        # === Construct output matrices for our economy using the solution to the LQ problem === #

        self.A0 = self.A - self.B.dot(self.F)

        self.Sh = self.A0[0:self.nh, 0:self.nx]
        self.Sk = self.A0[self.nh:self.nh + self.nk, 0:self.nx]
        self.Sk1 = np.hstack((np.zeros((self.nk, self.nh)), np.eye(
            self.nk), np.zeros((self.nk, self.nz))))
        self.Si = -self.F
        self.Sd = np.hstack((np.zeros((self.nd, self.nh + self.nk)), self.ud))
        self.Sb = np.hstack((np.zeros((self.nb, self.nh + self.nk)), self.ub))
        self.Sc = uc.dot(phiin).dot(-self.phii.dot(self.Si) +
                                    self.gamma.dot(self.Sk1) + self.Sd)
        self.Sg = ug.dot(phiin).dot(-self.phii.dot(self.Si) +
                                    self.gamma.dot(self.Sk1) + self.Sd)
        self.Ss = self.llambda.dot(np.hstack((np.eye(self.nh), np.zeros(
            (self.nh, self.nk + self.nz))))) + self.pih.dot(self.Sc)

        # ===  Calculate eigenvalues of A0 === #
        self.A110 = self.A0[0:self.nh + self.nk, 0:self.nh + self.nk]
        self.endo = np.linalg.eigvals(self.A110)
        self.exo = np.linalg.eigvals(self.a22)

        # === Construct matrices for Lagrange Multipliers === #

        self.Mk = -2 * self.beta.item() * (np.hstack((np.zeros((self.nk, self.nh)), np.eye(
            self.nk), np.zeros((self.nk, self.nz))))).dot(self.P).dot(self.A0)
        self.Mh = -2 * self.beta.item() * (np.hstack((np.eye(self.nh), np.zeros(
            (self.nh, self.nk)), np.zeros((self.nh, self.nz))))).dot(self.P).dot(self.A0)
        self.Ms = -(self.Sb - self.Ss)
        self.Md = -(np.linalg.inv(np.vstack((self.phic.T, self.phig.T))).dot(
            np.vstack((self.thetah.T.dot(self.Mh) + self.pih.T.dot(self.Ms), -self.Sg))))
        self.Mc = -(self.thetah.T.dot(self.Mh) + self.pih.T.dot(self.Ms))
        self.Mi = -(self.thetak.T.dot(self.Mk))

    def compute_steadystate(self, nnc=2):
        """
        Computes the non-stochastic steady-state of the economy.

        Parameters
        ----------
        nnc : array_like(float)
            nnc is the location of the constant in the state vector x_t

        """
        zx = np.eye(self.A0.shape[0])-self.A0
        self.zz = nullspace(zx)
        self.zz /= self.zz[nnc]
        self.css = self.Sc.dot(self.zz)
        self.sss = self.Ss.dot(self.zz)
        self.iss = self.Si.dot(self.zz)
        self.dss = self.Sd.dot(self.zz)
        self.bss = self.Sb.dot(self.zz)
        self.kss = self.Sk.dot(self.zz)
        self.hss = self.Sh.dot(self.zz)

    def compute_sequence(self, x0, ts_length=None, Pay=None):
        """
        Simulate quantities and prices for the economy

        Parameters
        ----------
        x0 : array_like(float)
            The initial state

        ts_length : scalar(int)
            Length of the simulation

        Pay : array_like(float)
            Vector to price an asset whose payout is Pay*xt

        """
        lq = LQ(self.Q, self.R, self.A, self.B,
                self.C, N=self.W, beta=self.beta)
        xp, up, wp = lq.compute_sequence(x0, ts_length)
        self.h = self.Sh.dot(xp)
        self.k = self.Sk.dot(xp)
        self.i = self.Si.dot(xp)
        self.b = self.Sb.dot(xp)
        self.d = self.Sd.dot(xp)
        self.c = self.Sc.dot(xp)
        self.g = self.Sg.dot(xp)
        self.s = self.Ss.dot(xp)

        # === Value of J-period risk-free bonds === #
        # === See p.145: Equation (7.11.2) === #
        e1 = np.zeros((1, self.nc))
        e1[0, 0] = 1
        self.R1_Price = np.empty((ts_length + 1, 1))
        self.R2_Price = np.empty((ts_length + 1, 1))
        self.R5_Price = np.empty((ts_length + 1, 1))
        for i in range(ts_length + 1):
            self.R1_Price[i, 0] = self.beta * e1.dot(self.Mc).dot(np.linalg.matrix_power(
                self.A0, 1)).dot(xp[:, i]) / e1.dot(self.Mc).dot(xp[:, i])
            self.R2_Price[i, 0] = self.beta**2 * e1.dot(self.Mc).dot(
                np.linalg.matrix_power(self.A0, 2)).dot(xp[:, i]) / e1.dot(self.Mc).dot(xp[:, i])
            self.R5_Price[i, 0] = self.beta**5 * e1.dot(self.Mc).dot(
                np.linalg.matrix_power(self.A0, 5)).dot(xp[:, i]) / e1.dot(self.Mc).dot(xp[:, i])

        # === Gross rates of return on 1-period risk-free bonds === #
        self.R1_Gross = 1 / self.R1_Price

        # === Net rates of return on J-period risk-free bonds === #
        # === See p.148: log of gross rate of return, divided by j === #
        self.R1_Net = np.log(1 / self.R1_Price) / 1
        self.R2_Net = np.log(1 / self.R2_Price) / 2
        self.R5_Net = np.log(1 / self.R5_Price) / 5

        # === Value of asset whose payout vector is Pay*xt === #
        # See p.145: Equation (7.11.1)
        if isinstance(Pay, np.ndarray) == True:
            self.Za = Pay.T.dot(self.Mc)
            self.Q = solve_discrete_lyapunov(
                self.A0.T * self.beta**0.5, self.Za)
            self.q = self.beta / (1 - self.beta) * \
                np.trace(self.C.T.dot(self.Q).dot(self.C))
            self.Pay_Price = np.empty((ts_length + 1, 1))
            self.Pay_Gross = np.empty((ts_length + 1, 1))
            self.Pay_Gross[0, 0] = np.nan
            for i in range(ts_length + 1):
                self.Pay_Price[i, 0] = (xp[:, i].T.dot(self.Q).dot(
                    xp[:, i]) + self.q) / e1.dot(self.Mc).dot(xp[:, i])
            for i in range(ts_length):
                self.Pay_Gross[i + 1, 0] = self.Pay_Price[i + 1,
                                                          0] / (self.Pay_Price[i, 0] - Pay.dot(xp[:, i]))
        return

    def irf(self, ts_length=100, shock=None):
        """
        Create Impulse Response Functions

        Parameters
        ----------

        ts_length : scalar(int)
            Number of periods to calculate IRF

        Shock : array_like(float)
            Vector of shocks to calculate IRF to. Default is first element of w

        """

        if type(shock) != np.ndarray:
            # Default is to select first element of w
            shock = np.vstack((np.ones((1, 1)), np.zeros((self.nw - 1, 1))))

        self.c_irf = np.empty((ts_length, self.nc))
        self.s_irf = np.empty((ts_length, self.nb))
        self.i_irf = np.empty((ts_length, self.ni))
        self.k_irf = np.empty((ts_length, self.nk))
        self.h_irf = np.empty((ts_length, self.nh))
        self.g_irf = np.empty((ts_length, self.ng))
        self.d_irf = np.empty((ts_length, self.nd))
        self.b_irf = np.empty((ts_length, self.nb))

        for i in range(ts_length):
            self.c_irf[i, :] = self.Sc.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T
            self.s_irf[i, :] = self.Ss.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T
            self.i_irf[i, :] = self.Si.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T
            self.k_irf[i, :] = self.Sk.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T
            self.h_irf[i, :] = self.Sh.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T
            self.g_irf[i, :] = self.Sg.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T
            self.d_irf[i, :] = self.Sd.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T
            self.b_irf[i, :] = self.Sb.dot(
                np.linalg.matrix_power(self.A0, i)).dot(self.C).dot(shock).T

        return

    def canonical(self):
        """
        Compute canonical preference representation
        Uses auxiliary problem of 9.4.2, with the preference shock process reintroduced
        Calculates pihat, llambdahat and ubhat for the equivalent canonical household technology
        """
        Ac1 = np.hstack((self.deltah, np.zeros((self.nh, self.nz))))
        Ac2 = np.hstack((np.zeros((self.nz, self.nh)), self.a22))
        Ac = np.vstack((Ac1, Ac2))
        Bc = np.vstack((self.thetah, np.zeros((self.nz, self.nc))))
        Rc1 = np.hstack((self.llambda.T.dot(self.llambda), -
                         self.llambda.T.dot(self.ub)))
        Rc2 = np.hstack((-self.ub.T.dot(self.llambda), self.ub.T.dot(self.ub)))
        Rc = np.vstack((Rc1, Rc2))
        Qc = self.pih.T.dot(self.pih)
        Nc = np.hstack(
            (self.pih.T.dot(self.llambda), -self.pih.T.dot(self.ub)))

        lq_aux = LQ(Qc, Rc, Ac, Bc, N=Nc, beta=self.beta)

        P1, F1, d1 = lq_aux.stationary_values()

        self.F_b = F1[:, 0:self.nh]
        self.F_f = F1[:, self.nh:]

        self.pihat = np.linalg.cholesky(self.pih.T.dot(
            self.pih) + self.beta.dot(self.thetah.T).dot(P1[0:self.nh, 0:self.nh]).dot(self.thetah)).T
        self.llambdahat = self.pihat.dot(self.F_b)
        self.ubhat = - self.pihat.dot(self.F_f)

        return



"""
Tests for approximation.py file (i.e. tauchen)

"""
import numpy as np
import pytest
from quantecon.markov import tauchen, rouwenhorst, discrete_var
from numpy.testing import assert_, assert_allclose, assert_raises
import scipy as sp

class TestTauchen:

    def setup_method(self):
        self.rho, self.sigma = np.random.rand(2)
        self.n = np.random.randint(3, 25)
        self.n_std = np.random.randint(5)
        self.tol = 1e-12
        self.mu = 0.

        mc = tauchen(self.n, self.rho, self.sigma, self.mu, self.n_std)
        self.x, self.P = mc.state_values, mc.P

    def teardown_method(self):
        del self.x
        del self.P

    def testStateCenter(self):
        for mu in [0., 1., -1.]:
            mu_expect = mu / (1 - self.rho)
            mc = tauchen(self.n, self.rho, self.sigma, mu, self.n_std)
            assert_allclose(mu_expect, np.mean(mc.state_values), atol=self.tol)

    def testShape(self):
        i, j = self.P.shape

        assert_(i == j)

    def testDim(self):
        dim_x = self.x.ndim
        dim_P = self.P.ndim

        assert_(dim_x == 1 and dim_P == 2)

    def test_transition_mat_row_sum_1(self):
        assert_allclose(np.sum(self.P, axis=1), 1, atol=self.tol)

    def test_positive_probs(self):
        assert_(np.all(self.P > -self.tol))

    def test_states_sum_0(self):
        assert_(abs(np.sum(self.x)) < self.tol)

    def test_old_tauchen_api_warning(self):
        # Test the warning
        with pytest.warns(UserWarning):
            # This will raise an error because `n` must be an int
            assert_raises(TypeError, tauchen, 4.0, self.rho, self.sigma,
                                self.mu, self.n_std)

class TestRouwenhorst:

    def setup_method(self):
        self.rho, self.sigma = np.random.uniform(0, 1, size=2)
        self.n = np.random.randint(3, 26)
        self.mu = np.random.randint(0, 11)
        self.tol = 1e-10

        with pytest.warns(UserWarning):
            mc = rouwenhorst(self.n, self.rho, self.sigma, self.mu)
        self.x, self.P = mc.state_values, mc.P

    def teardown_method(self):
        del self.x
        del self.P

    def testShape(self):
        i, j = self.P.shape

        assert_(i == j)

    def testDim(self):
        dim_x = self.x.ndim
        dim_P = self.P.ndim
        assert_(dim_x == 1 and dim_P == 2)

    def test_transition_mat_row_sum_1(self):
        assert_allclose(np.sum(self.P, axis=1), 1, atol=self.tol)

    def test_positive_probs(self):
        assert_(np.all(self.P > -self.tol))

    def test_states_sum_0(self):
        tol = self.tol + self.n*(self.mu/(1 - self.rho))
        assert_(abs(np.sum(self.x)) < tol)

    def test_control_case(self):
        n = 3; mu = 1; sigma = 0.5; rho = 0.8;
        with pytest.warns(UserWarning):
            mc_rouwenhorst = rouwenhorst(n, rho, sigma, mu)
        mc_rouwenhorst.x, mc_rouwenhorst.P = mc_rouwenhorst.state_values, mc_rouwenhorst.P
        sigma_y = np.sqrt(sigma**2 / (1-rho**2))
        psi = sigma_y * np.sqrt(n-1)
        known_x = np.array([-psi+5.0, 5., psi+5.0])
        known_P = np.array(
            [[0.81, 0.18, 0.01], [0.09, 0.82, 0.09], [0.01, 0.18, 0.81]])
        assert_(np.sum(mc_rouwenhorst.x - known_x) < self.tol and
                np.sum(mc_rouwenhorst.P - known_P) < self.tol)


class TestDiscreteVar:

    def setup_method(self):
        self.random_state = np.random.RandomState(0)

        Omega = np.array([[0.0012346, -0.0000776],
                          [-0.0000776, 0.0000401]])
        self.A = [[0.7901, -1.3570], 
                  [-0.0104, 0.8638]]
        self.C = sp.linalg.sqrtm(Omega)
        self.T= 1_000_000
        self.sizes = np.array((2, 3))

        # Expected outputs
        self.S_out = [
                     [-0.38556417, -0.05387746],
                     [-0.38556417,  0.        ],
                     [-0.38556417,  0.05387746],
                     [ 0.38556417, -0.05387746],
                     [ 0.38556417,  0.        ],
                     [ 0.38556417,  0.05387746]]

        self.S_out_orderF = [
                     [-0.38556417, -0.05387746],
                     [ 0.38556417, -0.05387746],
                     [-0.38556417,  0.        ],
                     [ 0.38556417,  0.        ],
                     [-0.38556417,  0.05387746],
                     [ 0.38556417,  0.05387746]]
            
        self.P_out = [
            [1.61290323e-02, 1.12903226e-01, 0.00000000e+00, 3.70967742e-01, 
            5.00000000e-01, 0.00000000e+00],
            [1.00964548e-04, 8.51048124e-01, 3.82857566e-02, 4.03858192e-04,
            1.10111936e-01, 4.93604457e-05],
            [0.00000000e+00, 3.02295449e-01, 6.97266822e-01, 0.00000000e+00,
            3.85201268e-04, 5.25274456e-05],
            [3.60600761e-05, 4.86811027e-04, 0.00000000e+00, 6.97473992e-01,
            3.02003137e-01, 0.00000000e+00],
            [3.17039037e-05, 1.11090478e-01, 4.55177474e-04, 3.75374219e-02,
            8.50778784e-01, 1.06434534e-04],
            [0.00000000e+00, 4.45945946e-01, 3.37837838e-01, 0.00000000e+00,
            1.89189189e-01, 2.70270270e-02]]

        self.P_out_orderF =[
            [1.61290323e-02, 3.70967742e-01, 1.12903226e-01, 5.00000000e-01, 
            0.00000000e+00, 0.00000000e+00],
            [3.60600761e-05, 6.97473992e-01, 4.86811027e-04, 3.02003137e-01, 
            0.00000000e+00, 0.00000000e+00],
            [1.00964548e-04, 4.03858192e-04, 8.51048124e-01, 1.10111936e-01, 
            3.82857566e-02, 4.93604457e-05],
            [3.17039037e-05, 3.75374219e-02, 1.11090478e-01, 8.50778784e-01, 
            4.55177474e-04, 1.06434534e-04],
            [0.00000000e+00, 0.00000000e+00, 3.02295449e-01, 3.85201268e-04, 
            6.97266822e-01, 5.25274456e-05],
            [0.00000000e+00, 0.00000000e+00, 4.45945946e-01, 1.89189189e-01, 
            3.37837838e-01, 2.70270270e-02]]
        
        self.P_out_non_square = [
            [3.70370370e-02, 2.22222222e-01, 0.00000000e+00, 4.25925926e-01, 
            3.14814815e-01, 0.00000000e+00],
            [6.92865939e-05, 8.50870664e-01, 3.83177215e-02, 4.17954615e-04, 
            1.10275202e-01, 4.91711312e-05],
            [0.00000000e+00, 3.08160000e-01, 6.91360000e-01, 0.00000000e+00, 
            3.91111111e-04, 8.88888889e-05],
            [1.08405001e-04, 5.05890005e-04, 0.00000000e+00, 6.95707162e-01, 
            3.03678543e-01, 0.00000000e+00],
            [3.40242525e-05, 1.11864937e-01, 4.44583566e-04, 3.77260912e-02, 
            8.49844169e-01, 8.61947730e-05],
            [0.00000000e+00, 4.70588235e-01, 3.08823529e-01, 0.00000000e+00, 
            1.76470588e-01, 4.41176471e-02]]

        self.A, self.C, self.S_out, self.P_out, self.S_out_orderF,\
        self.P_out_orderF, self.P_out_non_square \
            = map(np.array,(self.A, self.C, self.S_out, self.P_out, 
                                self.S_out_orderF, self.P_out_orderF, 
                                self.P_out_non_square))

    def teardown_method(self):
        del self.A, self.C, self.S_out, self.P_out

    def test_discretization(self):
        mc = discrete_var(
                self.A, self.C, grid_sizes=self.sizes,
                sim_length=self.T, random_state=self.random_state)
        assert_allclose(mc.state_values, self.S_out)
        assert_allclose(mc.P, self.P_out)

    def test_sp_distributions(self):
        mc = discrete_var(
                self.A, self.C, 
                grid_sizes=self.sizes,
                sim_length=self.T, 
                rv=sp.stats.multivariate_normal(cov=np.identity(2)),
                random_state=self.random_state)
        assert_allclose(mc.state_values, self.S_out)
        assert_allclose(mc.P, self.P_out)

    def test_order_F(self):
        mc = discrete_var(
                self.A, self.C, 
                grid_sizes=self.sizes,
                sim_length=self.T, 
                order='F',
                random_state=self.random_state)
        assert_allclose(mc.state_values, self.S_out_orderF)
        assert_allclose(mc.P, self.P_out_orderF)

    def test_order_non_squareC(self):
        new_col = np.array([0, 0])
        self.C = np.insert(self.C, 2, new_col, axis=1)
        mc = discrete_var(
            self.A, self.C, 
            grid_sizes=self.sizes,
            sim_length=self.T, 
            order='C',
            random_state=self.random_state)
        assert_allclose(mc.state_values, self.S_out)
        assert_allclose(mc.P, self.P_out_non_square)



"""
Tests for gridtools.py file

"""
import numpy as np
import time
import pytest
from numpy.testing import (
    assert_array_equal, assert_equal, assert_, assert_raises
)

from quantecon._gridtools import (
    cartesian, mlinspace, _repeat_1d, simplex_grid, simplex_index,
    num_compositions, num_compositions_jit, cartesian_nearest_index
)


def test_cartesian_C_order():
    x = np.linspace(0, 9, 10)

    prod = cartesian([x, x, x])

    correct = True
    for i in range(999):
        n = prod[i, 0]*100+prod[i, 1]*10+prod[i, 2]
        correct *= (i == n)

    assert_(correct)


def test_cartesian_C_order_int_float():
    x_int = np.arange(10)
    x_float = np.linspace(0, 9, 10)
    prod_int = cartesian([x_int]*3)
    prod_float = cartesian([x_float]*3)
    assert_(prod_int.dtype == x_int.dtype)
    assert_(prod_float.dtype == x_float.dtype)
    assert_(abs(prod_int-prod_float).max() == 0)


def test_cartesian_C_order_int_float_mixed():
    x_int = [0, 1]
    x_float = [2.3, 4.5]
    prod_expected = np.array(
        [[0., 2.3],
         [0., 4.5],
         [1., 2.3],
         [1., 4.5]]
    )
    prod_computed = cartesian([x_int, x_float])
    assert_array_equal(prod_computed, prod_expected)


def test_cartesian_F_order():
    x = np.linspace(0, 9, 10)

    prod = cartesian([x, x, x], order='F')

    correct = True
    for i in range(999):
        n = prod[i, 2]*100+prod[i, 1]*10+prod[i, 0]
        correct *= (i == n)

    assert_(correct)


@pytest.mark.slow
def test_performance_C():
    N_x = 1000
    N_y = 7777
    x = np.linspace(1, N_x, N_x)
    y = np.linspace(1, N_y, N_y)

    cartesian([x[:10], y[:10]]) # warmup

    t1 = time.time()
    for i in range(100):
        prod = cartesian([x, y])
    t2 = time.time()
    # print(prod.shape)

    # compute the same produce using numpy:

    t3 = time.time()
    for i in range(100):
        prod_numpy = np.column_stack([
            np.repeat(x, N_y),
            np.tile(y, N_x)
        ])
    t4 = time.time()

    print("Timings for 'cartesian' (C order)")
    print("Cartesian: {}".format(t2-t1))
    print("Numpy:     {}".format(t4-t3))
    assert_(abs(prod-prod_numpy).max() == 0)


@pytest.mark.slow
def test_performance_F():
    N_x = 1000
    N_y = 7777
    x = np.linspace(1, N_x, N_x)
    y = np.linspace(1, N_y, N_y)

    cartesian([x[:10], y[:10]]) # warmup

    t1 = time.time()
    for i in range(100):
        prod = cartesian([x, y], order='F')
    t2 = time.time()
    # print(prod.shape)

    # compute the same produce using numpy:

    t3 = time.time()
    for i in range(100):
        prod_numpy = np.column_stack([
            np.tile(x, N_y),
            np.repeat(y, N_x)
        ])
    t4 = time.time()

    print("Timings for 'cartesian'(Fortran order)")
    print("Cartesian: {}".format(t2-t1))
    print("Numpy:     {}".format(t4-t3))
    assert_(abs(prod-prod_numpy).max() == 0)


def test_mlinsplace():
    mlinspace([-1, -1], [2, 3], [30, 50])
    cartesian([np.linspace(-1, 2, 30), np.linspace(-1, 3, 50)])


def test_tile():
    x = np.linspace(1, 100, 100)
    t1 = time.time()
    t_repeat = np.zeros(100*1000)
    _repeat_1d(x, 1, t_repeat)
    t2 = time.time()

    t3 = time.time()
    t_numpy = np.tile(x, 1000)
    t4 = time.time()

    print("Timings for 'tile' operation")
    print("Repeat_1d: {}".format(t2-t1))
    print("Numpy:     {}".format(t4-t3))

    assert_(abs(t_numpy-t_repeat).max())


def test_repeat():
    x = np.linspace(1, 100, 100)
    t1 = time.time()
    t_repeat = np.zeros(100*1000)
    _repeat_1d(x, 1000, t_repeat)
    t2 = time.time()

    t3 = time.time()
    t_numpy = np.repeat(x, 1000)
    t4 = time.time()

    print("Timings for 'repeat' operation")
    print("Repeat_1d: {}".format(t2-t1))
    print("Numpy:     {}".format(t4-t3))

    assert_(abs(t_numpy-t_repeat).max())


class TestCartesianNearestIndex:
    def setup_method(self):
        nums = (5, 6)
        self.nodes = [list(range(nums[0])), np.linspace(0, 1, nums[1])]
        self.orders = ['C', 'F']
        self.prod_dict = \
            {order:cartesian(self.nodes, order=order) for order in self.orders}

    def linear_search(self, x, order='C'):
        x = np.asarray(x)
        return ((self.prod_dict[order] - x)**2).sum(1).argmin()

    def test_1d(self):
        x = (1.2, 0.3)
        for order in self.orders:
            ind_expected = self.linear_search(x, order)
            ind_computed = cartesian_nearest_index(x, self.nodes, order)
            assert_equal(ind_computed, ind_expected)

        assert_raises(
            ValueError, cartesian_nearest_index, x, self.prod_dict['C']
        )

    def test_2d(self):
        T = 10
        rng = np.random.default_rng(1234)
        X = np.column_stack((
            rng.uniform(self.nodes[0][0]-1, self.nodes[0][-1]+1, size=T),
            rng.standard_normal(T) + 0.5
        ))
        ind_expected = np.empty(T, dtype=np.intp)

        for order in self.orders:
            for t in range(T):
                ind_expected[t] = self.linear_search(X[t], order)
            ind_computed = cartesian_nearest_index(X, self.nodes, order)
            assert_array_equal(ind_computed, ind_expected)

        assert_raises(
            ValueError, cartesian_nearest_index, X, self.prod_dict['C']
        )


class TestSimplexGrid:
    def setup_method(self):
        self.simplex_grid_3_4 = np.array([[0, 0, 4],
                                          [0, 1, 3],
                                          [0, 2, 2],
                                          [0, 3, 1],
                                          [0, 4, 0],
                                          [1, 0, 3],
                                          [1, 1, 2],
                                          [1, 2, 1],
                                          [1, 3, 0],
                                          [2, 0, 2],
                                          [2, 1, 1],
                                          [2, 2, 0],
                                          [3, 0, 1],
                                          [3, 1, 0],
                                          [4, 0, 0]])

    def test_simplex_grid(self):
        out = simplex_grid(3, 4)
        assert_array_equal(out, self.simplex_grid_3_4)

        assert_array_equal(simplex_grid(1, 1), [[1]])

    def test_simplex_index(self):
        points = [[0, 0, 4], [1, 1, 2], [4, 0, 0]]
        for point in points:
            idx = simplex_index(point, 3, 4)
            assert_array_equal(self.simplex_grid_3_4[idx], point)

        assert_(simplex_index([1], 1, 1) == 0)

    def test_num_compositions(self):
        num = num_compositions(3, 4)
        assert_(num == len(self.simplex_grid_3_4))

    def test_num_compositions_jit(self):
        num = num_compositions_jit(3, 4)
        assert_(num == len(self.simplex_grid_3_4))

        # Exceed max value of np.intp
        assert_(num_compositions_jit(100, 50) == 0)


def test_simplex_grid_raises_value_error_overflow():
    # Exceed max value of np.intp
    assert_raises(ValueError, simplex_grid, 100, 50)



"""
Contain a minmax problem solver routine.

"""
import numpy as np
from numba import jit
from .linprog_simplex import solve_tableau, PivOptions
from .pivoting import _pivoting


@jit(nopython=True, cache=True)
def minmax(A, max_iter=10**6, piv_options=PivOptions()):
    r"""
    Given an m x n matrix `A`, return the value :math:`v^*` of the
    minmax problem:

    .. math::

        v^* = \max_{x \in \Delta_m} \min_{y \in \Delta_n} x^T A y
            = \min_{y \in \Delta_n}\max_{x \in \Delta_m} x^T A y

    and the optimal solutions :math:`x^* \in \Delta_m` and
    :math:`y^* \in \Delta_n`: :math:`v^* = x^{*T} A y^*`, where
    :math:`\Delta_k = \{z \in \mathbb{R}^k_+ \mid z_1 + \cdots + z_k =
    1\}`, :math:`k = m, n`.

    This routine is jit-compiled by Numba, using
    `optimize.linprog_simplex` routines.

    Parameters
    ----------
    A : ndarray(float, ndim=2)
        ndarray of shape (m, n).

    max_iter : int, optional(default=10**6)
        Maximum number of iteration in the linear programming solver.

    piv_options : PivOptions, optional
        PivOptions namedtuple to set tolerance values used in the linear
        programming solver.

    Returns
    -------
    v : float
        Value :math:`v^*` of the minmax problem.

    x : ndarray(float, ndim=1)
        Optimal solution :math:`x^*`, of shape (m,).

    y : ndarray(float, ndim=1)
        Optimal solution :math:`y^*`, of shape (n,).

    """
    m, n = A.shape

    min_ = A.min()
    const = 0.
    if min_ <= 0:
        const = min_ * (-1) + 1

    tableau = np.zeros((m+2, n+1+m+1))

    for i in range(m):
        for j in range(n):
            tableau[i, j] = A[i, j] + const
        tableau[i, n] = -1
        tableau[i, n+1+i] = 1

    tableau[-2, :n] = 1
    tableau[-2, -1] = 1
    tableau[-1, n] = -1

    # Phase 1
    pivcol = 0

    pivrow = 0
    max_ = tableau[0, pivcol]
    for i in range(1, m):
        if tableau[i, pivcol] > max_:
            pivrow = i
            max_ = tableau[i, pivcol]

    _pivoting(tableau, n, pivrow)
    _pivoting(tableau, pivcol, m)

    basis = np.arange(n+1, n+1+m+1)
    basis[pivrow] = n
    basis[-1] = 0

    # Phase 2
    solve_tableau(tableau, basis, max_iter-2, skip_aux=False,
                  piv_options=piv_options)

    # Obtain solution
    x = np.empty(m)
    y = np.zeros(n)

    for i in range(m+1):
        if basis[i] < n:
            y[basis[i]] = tableau[i, -1]

    for j in range(m):
        x[j] = tableau[-1, n+1+j]
        if x[j] != 0:
            x[j] *= -1

    v = tableau[-1, -1] - const

    return v, x, y



"""
Tests for dle.py file
"""

import numpy as np
from numpy.testing import assert_allclose
from quantecon import DLE

ATOL = 1e-10


class TestDLE:

    def setup_method(self):
        """
        Given LQ control is tested we will test the transformation
        to alter the problem into a form suitable to solve using LQ
        """
        # Initial Values
        gam = 0
        gamma = np.array([[gam], [0]])
        phic = np.array([[1], [0]])
        phig = np.array([[0], [1]])
        phi1 = 1e-4
        phii = np.array([[0], [-phi1]])
        deltak = np.array([[.95]])
        thetak = np.array([[1]])
        beta = np.array([[1 / 1.05]])
        ud = np.array([[5, 1, 0], [0, 0, 0]])
        a22 = np.array([[1, 0, 0], [0, 0.8, 0], [0, 0, 0.5]])
        c2 = np.array([[0, 1, 0], [0, 0, 1]]).T
        llambda = np.array([[0]])
        pih = np.array([[1]])
        deltah = np.array([[.9]])
        thetah = np.array([[1]]) - deltah
        ub = np.array([[30, 0, 0]])

        information = (a22, c2, ub, ud)
        technology = (phic, phig, phii, gamma, deltak, thetak)
        preferences = (beta, llambda, pih, deltah, thetah)

        self.dle = DLE(information, technology, preferences)

    def teardown_method(self):
        del self.dle

    def test_transformation_Q(self):
        Q_solution = np.array([[5.e-09]])
        assert_allclose(Q_solution, self.dle.Q)

    def test_transformation_R(self):
        R_solution = np.array([[0.,   0.,   0.,   0.,   0.],
                               [0.,   0.,   0.,   0.,   0.],
                               [0.,   0., 312.5, -12.5,   0.],
                               [0.,   0., -12.5,   0.5,   0.],
                               [0.,   0.,   0.,   0.,   0.]])
        assert_allclose(R_solution, self.dle.R)

    def test_transformation_A(self):
        A_solution = np.array([[0.9, 0., 0.5, 0.1, 0.],
                               [0., 0.95, 0., 0., 0.],
                               [0., 0., 1., 0., 0.],
                               [0., 0., 0., 0.8, 0.],
                               [0., 0., 0., 0., 0.5]])
        assert_allclose(A_solution, self.dle.A)

    def test_transformation_B(self):
        B_solution = np.array([[-0.],
                               [1.],
                               [0.],
                               [0.],
                               [0.]])
        assert_allclose(B_solution, self.dle.B)

    def test_transformation_C(self):
        C_solution = np.array([[0., 0.],
                               [0., 0.],
                               [0., 0.],
                               [1., 0.],
                               [0., 1.]])
        assert_allclose(C_solution, self.dle.C)

    def test_transformation_W(self):
        W_solution = np.array([[0., 0., 0., 0., 0.]])
        assert_allclose(W_solution, self.dle.W)

    def test_compute_steadystate(self):
        solutions = {
            'css' : np.array([[5.]]),
            'sss' : np.array([[5.]]),
            'iss' : np.array([[0.]]),
            'dss' : np.array([[5.], [0.]]),
            'bss' : np.array([[30.]]),
            'kss' : np.array([[0.]]),
            'hss' : np.array([[5.]]),
        }
        self.dle.compute_steadystate()
        for item in solutions.keys():
            assert_allclose(self.dle.__dict__[
                            item], solutions[item], atol=ATOL)

    def test_canonical(self):
        solutions = {
            'pihat': np.array([[1.]]),
            'llambdahat': np.array([[-1.48690584e-19]]),
            'ubhat': np.array([[30., -0., -0.]])
        }
        self.dle.canonical()
        for item in solutions.keys():
            assert_allclose(self.dle.__dict__[
                            item], solutions[item], atol=ATOL)



"""
Tests for robustlq.py

"""
import numpy as np
from numpy.testing import assert_allclose, assert_
from quantecon import LQ
from quantecon import RBLQ


class TestRBLQControl:

    def setup_method(self):
        # Initial Values
        a_0     = 100
        a_1     = 0.5
        rho     = 0.9
        sigma_d = 0.05
        beta    = 0.95
        c       = 2
        gamma   = 50.0
        theta = 0.002
        ac    = (a_0 - c) / 2.0

        R = np.array([[0,  ac,    0],
                      [ac, -a_1, 0.5],
                      [0., 0.5,  0]])

        R = -R
        Q = gamma / 2
        Q_pf = 0.

        A = np.array([[1., 0., 0.],
                      [0., 1., 0.],
                      [0., 0., rho]])
        B = np.array([[0.],
                      [1.],
                      [0.]])
        B_pf = np.zeros((3, 1))

        C = np.array([[0.],
                      [0.],
                      [sigma_d]])

        # the *_pf endings refer to an example with pure forecasting
        # (see p171 in Robustness)
        self.rblq_test = RBLQ(Q, R, A, B, C, beta, theta)
        self.rblq_test_pf = RBLQ(Q_pf, R, A, B_pf, C, beta, theta)
        self.lq_test = LQ(Q, R, A, B, C, beta=beta)
        self.methods = ['doubling', 'qz']

    def teardown_method(self):
        del self.rblq_test
        del self.rblq_test_pf

    def test_pure_forecasting(self):
        assert_(self.rblq_test_pf.pure_forecasting)

    def test_robust_rule_vs_simple(self):
        rblq = self.rblq_test
        rblq_pf = self.rblq_test_pf

        for method in self.methods:
            Fr, Kr, Pr = self.rblq_test.robust_rule(method=method)
            Fr_pf, Kr_pf, Pr_pf = self.rblq_test_pf.robust_rule(method=method)

            Fs, Ks, Ps = rblq.robust_rule_simple(P_init=Pr, tol=1e-12)
            Fs_pf, Ks_pf, Ps_pf = rblq_pf.robust_rule_simple(
                P_init=Pr_pf, tol=1e-12)

            assert_allclose(Fr, Fs, rtol=1e-4)
            assert_allclose(Kr, Ks, rtol=1e-4)
            assert_allclose(Pr, Ps, rtol=1e-4)

            atol = 1e-10
            assert_allclose(Fr_pf, Fs_pf, rtol=1e-4)
            assert_allclose(Kr_pf, Ks_pf, rtol=1e-4, atol=atol)
            assert_allclose(Pr_pf, Ps_pf, rtol=1e-4, atol=atol)


    def test_f2k_and_k2f(self):
        rblq = self.rblq_test

        for method in self.methods:
            Fr, Kr, Pr = self.rblq_test.robust_rule(method=method)
            K_f2k, P_f2k = rblq.F_to_K(Fr, method=method)
            F_k2f, P_k2f = rblq.K_to_F(Kr, method=method)
            assert_allclose(K_f2k, Kr, rtol=1e-4)
            assert_allclose(F_k2f, Fr, rtol=1e-4)
            assert_allclose(P_f2k, P_k2f, rtol=1e-4)

    def test_evaluate_F(self):
        rblq = self.rblq_test
        for method in self.methods:
            Fr, Kr, Pr = self.rblq_test.robust_rule(method=method)

            Kf, Pf, df, Of, of = rblq.evaluate_F(Fr)

            # In the future if we wanted, we could check more things, but I
            # think the other pieces are basically just plugging these into
            # equations so if these hold then the others should be correct
            # as well.
            assert_allclose(Pf, Pr)
            assert_allclose(Kf, Kr)



# flake8: noqa
"""
Random SubPackage
=================

Utilities
---------
Utilities to Support Generation of Random Arrays or Matrices
1. probvec
2. sample_without_replacement

.. Future Work
   -----------
   1. AR1 Function
"""

from .utilities import probvec, sample_without_replacement, draw



"""
Tests for ivp.py

"""

import numpy as np

from quantecon import IVP

# use the Solow Model with Cobb-Douglas production as test case
def solow_model(t, k, g, n, s, alpha, delta):
    r"""
    Equation of motion for capital stock (per unit effective labor).

    Parameters
    ----------
    t : float
        Time
    k : ndarray (float, shape=(1,))
        Capital stock (per unit of effective labor)
    g : float
        Growth rate of technology.
    n : float
        Growth rate of the labor force.
    s : float
        Savings rate. Must satisfy `0 < s < 1`.
    alpha : float
        Elasticity of output with respect to capital stock. Must satisfy
        :math:`0 < alpha < 1`.
    delta : float
        Depreciation rate of physical capital. Must satisfy :math:`0 < \delta`.

    Returns
    -------
    k_dot : ndarray (float, shape(1,))
        Time derivative of capital stock (per unit effective labor).

    """
    k_dot = s * k**alpha - (g + n + delta) * k
    return k_dot


def solow_jacobian(t, k, g, n, s, alpha, delta):
    r"""
    Jacobian matrix for the Solow model.

    Parameters
    ----------
    t : float
        Time
    k : ndarray (float, shape=(1,))
        Capital stock (per unit of effective labor)
    g : float
        Growth rate of technology.
    n : float
        Growth rate of the labor force.
    s : float
        Savings rate. Must satisfy `0 < s < 1`.
    alpha : float
        Elasticity of output with respect to capital stock. Must satisfy
        :math:`0 < alpha < 1`.
    delta : float
        Depreciation rate of physical capital. Must satisfy :math:`0 < \delta`.

    Returns
    -------
    jac : ndarray (float, shape(1,))
        Time derivative of capital stock (per unit effective labor).

    """
    jac = s * alpha * k**(alpha - 1) - (g + n + delta)
    return jac


def solow_steady_state(g, n, s, alpha, delta):
    r"""
    Steady-state level of capital stock (per unit effective labor).

    Parameters
    ----------
    g : float
        Growth rate of technology.
    n : float
        Growth rate of the labor force.
    s : float
        Savings rate. Must satisfy `0 < s < 1`.
    alpha : float
        Elasticity of output with respect to capital stock. Must satisfy
        :math:`0 < alpha < 1`.
    delta : float
        Depreciation rate of physical capital. Must satisfy :math:`0 < \delta`.

    Returns
    -------
    kstar : float
        Steady state value of capital stock (per unit effective labor).

    """
    k_star = (s / (n + g + delta))**(1 / (1 - alpha))
    return k_star


def solow_analytic_solution(t, k0, g, n, s, alpha, delta):
    r"""
    Analytic solution for the path of capital stock (per unit effective labor).

    Parameters
    ----------
    t : ndarray(float, shape=(1,))
        Time
    k : ndarray (float, shape=(1,))
        Capital stock (per unit of effective labor)
    g : float
        Growth rate of technology.
    n : float
        Growth rate of the labor force.
    s : float
        Savings rate. Must satisfy `0 < s < 1`.
    alpha : float
        Elasticity of output with respect to capital stock. Must satisfy
        :math:`0 < alpha < 1`.
    delta : float
        Depreciation rate of physical capital. Must satisfy :math:`0 < \delta`.

    Returns
    -------
    soln : ndarray (float, shape(t.size, 2))
        Trajectory describing the analytic solution of the model.

    """
    # lambda governs the speed of convergence
    lmbda = (n + g + delta) * (1 - alpha)

    # analytic solution for Solow model at time t
    k_t = (((s / (n + g + delta)) * (1 - np.exp(-lmbda * t)) +
            k0**(1 - alpha) * np.exp(-lmbda * t))**(1 / (1 - alpha)))

    # combine into a (t.size, 2) array
    analytic_traj = np.hstack((t[:, np.newaxis], k_t[:, np.newaxis]))

    return analytic_traj

# create an instance of the IVP class
valid_params = (0.02, 0.02, 0.15, 0.33, 0.05)
model = IVP(f=solow_model,
                jac=solow_jacobian)

model.f_params = valid_params
model.jac_params = valid_params


# helper functions for conducting tests
def _compute_fixed_length_solns(model, t0, k0):
    """Returns a dictionary of fixed length solution trajectories."""

    # storage containter for integration results
    results = {}

    for integrator in ['dopri5', 'dop853', 'vode', 'lsoda']:

        # tighten tolerances so tests don't fail due to numerical issues
        discrete_soln = model.solve(t0, k0, h=1e0, T=1e3,
                                    integrator=integrator,
                                    atol=1e-14, rtol=1e-11)

        # store the result
        results[integrator] = discrete_soln

    return results


def _termination_condition(t, k, g, n, s, alpha, delta):
        """Terminate solver when we get close to steady state."""
        diff = k - solow_steady_state(g, n, s, alpha, delta)
        return diff


def _compute_variable_length_solns(model, t0, k0, g, tol):
    """Returns a dictionary of variable length solution trajectories."""

    # storage containter for integration results
    results = {}

    for integrator in ['dopri5', 'dop853', 'vode', 'lsoda']:

        # tighten tolerances so tests don't fail due to numerical issues
        discrete_soln = model.solve(t0, k0, h=1e0, g=g, tol=tol,
                                    integrator=integrator,
                                    atol=1e-14, rtol=1e-11)

        # store the result
        results[integrator] = discrete_soln

    return results


# testing functions
def test_solve_args():
    """Testing arguments passed to the IVP.solve method."""
    # g and tol must be passed together!
    with np.testing.assert_raises(ValueError):
        t0, k0 = 0, np.array([5.0])
        model.solve(t0, k0, g=_termination_condition)


def test_solve_fixed_trajectory():
    """Testing computation of fixed length solution trajectory."""

    # compute some fixed length trajectories
    t0, k0 = 0, np.array([5.0])
    results = _compute_fixed_length_solns(model, t0, k0)

    # conduct the tests
    for integrator, numeric_solution in results.items():
        ti = numeric_solution[:, 0]
        analytic_solution = solow_analytic_solution(ti, k0, *valid_params)
        np.testing.assert_allclose(numeric_solution, analytic_solution)


def test_solve_variable_trajectory():
    """Testing computation of variable length solution trajectory."""

    # compute some variable length trajectories
    t0, k0, tol = 0, np.array([5.0]), 1e-3
    results = _compute_variable_length_solns(model, t0, k0,
                                             g=_termination_condition, tol=tol)

    # conduct the tests
    for integrator, numeric_solution in results.items():
        ti = numeric_solution[:, 0]
        analytic_solution = solow_analytic_solution(ti, k0, *valid_params)

        # test accuracy of solution
        np.testing.assert_allclose(numeric_solution, analytic_solution)

        # test termination condition
        diff = numeric_solution[-1, 1] - solow_steady_state(*valid_params)
        np.testing.assert_(diff <= tol)


def test_interpolation():
    """Testing parametric B-spline interpolation methods."""

    # compute some fixed length trajectories
    t0, k0 = 0, np.array([5.0])
    results = _compute_fixed_length_solns(model, t0, k0)

    # conduct the tests
    for integrator, numeric_solution in results.items():

        # define an array of interpolation points
        N, T = 1000, numeric_solution[:, 0][-1]
        ti = np.linspace(t0, T, N)

        # used highest order B-spline interpolation available
        interp_solution = model.interpolate(numeric_solution, ti, k=3, ext=2)

        analytic_solution = solow_analytic_solution(ti, k0, *valid_params)
        np.testing.assert_allclose(interp_solution, analytic_solution)


def test_compute_residual():
    """Testing computation of solution residual."""

    # compute some fixed length trajectories
    t0, k0 = 0, np.array([5.0])
    results = _compute_fixed_length_solns(model, t0, k0)

    # conduct the tests
    for integrator, numeric_solution in results.items():

        # define an array of interpolation points
        N, T = 1000, numeric_solution[:, 0][-1]
        tmp_grid_pts = np.linspace(t0, T, N)

        # used highest order B-spline interpolation available
        tmp_residual = model.compute_residual(numeric_solution,
                                              tmp_grid_pts,
                                              k=5)

        expected_residual = np.zeros((N, 2))
        actual_residual = tmp_residual
        np.testing.assert_almost_equal(expected_residual, actual_residual)



"""
Tests for filter.py.
Using the data of original paper.

"""

import os
import pandas as pd
import numpy as np
from numpy.testing import assert_allclose
from quantecon import hamilton_filter
from quantecon.tests.util import get_data_dir

def test_hamilton_filter():
    # read data
    data_dir = get_data_dir()
    data = pd.read_csv(os.path.join(data_dir, "employment.csv"),
                       names = ['year', 'employment', 'matlab_cyc', 'matlab_cyc_rw'])

    data['hamilton_cyc'], data['hamilton_trend'] =  hamilton_filter(
        100*np.log(data['employment']), 8, 4)
    data['hamilton_cyc_rw'], data['hamilton_trend_rw'] = hamilton_filter(
        100*np.log(data['employment']), 8)
    assert_allclose(data['matlab_cyc'], data['hamilton_cyc'],
                    rtol = 1e-07, atol = 1e-07)
    assert_allclose(data['matlab_cyc_rw'], data['hamilton_cyc_rw'])



"""
Functions for working with periodograms of scalar data.

"""
import numpy as np
from numpy.fft import fft


def smooth(x, window_len=7, window='hanning'):
    """
    Smooth the data in x using convolution with a window of requested
    size and type.

    Parameters
    ----------
    x : array_like(float)
        A flat NumPy array containing the data to smooth
    window_len : scalar(int), optional
        An odd integer giving the length of the window.  Defaults to 7.
    window : string
        A string giving the window type. Possible values are 'flat',
        'hanning', 'hamming', 'bartlett' or 'blackman'

    Returns
    -------
    array_like(float)
        The smoothed values

    Notes
    -----
    Application of the smoothing window at the top and bottom of x is
    done by reflecting x around these points to extend it sufficiently
    in each direction.

    """
    if len(x) < window_len:
        raise ValueError("Input vector length must be >= window length.")

    if window_len < 3:
        raise ValueError("Window length must be at least 3.")

    if not window_len % 2:  # window_len is even
        window_len += 1
        print("Window length reset to {}".format(window_len))

    windows = {'hanning': np.hanning,
               'hamming': np.hamming,
               'bartlett': np.bartlett,
               'blackman': np.blackman,
               'flat': np.ones  # moving average
               }

    # === Reflect x around x[0] and x[-1] prior to convolution === #
    k = int(window_len / 2)
    xb = x[:k]   # First k elements
    xt = x[-k:]  # Last k elements
    s = np.concatenate((xb[::-1], x, xt[::-1]))

    # === Select window values === #
    if window in windows.keys():
        w = windows[window](window_len)
    else:
        msg = "Unrecognized window type '{}'".format(window)
        print(msg + " Defaulting to hanning")
        w = windows['hanning'](window_len)

    return np.convolve(w / w.sum(), s, mode='valid')


def periodogram(x, window=None, window_len=7):
    r"""
    Computes the periodogram

    .. math::

        I(w) = \frac{1}{n} \Big[ \sum_{t=0}^{n-1} x_t e^{itw} \Big] ^2

    at the Fourier frequencies :math:`w_j := \frac{2 \pi j}{n}`,
    :math:`j = 0, \dots, n - 1`, using the fast Fourier transform. Only the
    frequencies :math:`w_j` in :math:`[0, \pi]` and corresponding values
    :math:`I(w_j)` are returned. If a window type is given then smoothing
    is performed.

    Parameters
    ----------
    x : array_like(float)
        A flat NumPy array containing the data to smooth
    window_len : scalar(int), optional(default=7)
        An odd integer giving the length of the window.  Defaults to 7.
    window : string
        A string giving the window type. Possible values are 'flat',
        'hanning', 'hamming', 'bartlett' or 'blackman'

    Returns
    -------
    w : array_like(float)
        Fourier frequencies at which periodogram is evaluated
    I_w : array_like(float)
        Values of periodogram at the Fourier frequencies

    """
    n = len(x)
    I_w = np.abs(fft(x))**2 / n
    w = 2 * np.pi * np.arange(n) / n  # Fourier frequencies
    w, I_w = w[:int(n/2)+1], I_w[:int(n/2)+1]  # Take only values on [0, pi]
    if window:
        I_w = smooth(I_w, window_len=window_len, window=window)
    return w, I_w


def ar_periodogram(x, window='hanning', window_len=7):
    """
    Compute periodogram from data x, using prewhitening, smoothing and
    recoloring.  The data is fitted to an AR(1) model for prewhitening,
    and the residuals are used to compute a first-pass periodogram with
    smoothing.  The fitted coefficients are then used for recoloring.

    Parameters
    ----------
    x : array_like(float)
        A flat NumPy array containing the data to smooth
    window_len : scalar(int), optional
        An odd integer giving the length of the window.  Defaults to 7.
    window : string
        A string giving the window type. Possible values are 'flat',
        'hanning', 'hamming', 'bartlett' or 'blackman'

    Returns
    -------
    w : array_like(float)
        Fourier frequencies at which periodogram is evaluated
    I_w : array_like(float)
        Values of periodogram at the Fourier frequencies

    """
    # === run regression === #
    x_lag = x[:-1]  # lagged x
    X = np.array([np.ones(len(x_lag)), x_lag]).T  # add constant

    y = np.array(x[1:])  # current x

    beta_hat = np.linalg.solve(X.T @ X, X.T @ y)  # solve for beta hat
    e_hat = y - X @ beta_hat  # compute residuals
    phi = beta_hat[1]  # pull out phi parameter

    # === compute periodogram on residuals === #
    w, I_w = periodogram(e_hat, window=window, window_len=window_len)

    # === recolor and return === #
    I_w = I_w / np.abs(1 - phi * np.exp(1j * w))**2

    return w, I_w



"""
Implements the Kalman filter for a linear Gaussian state space model.

References
----------

https://lectures.quantecon.org/py/kalman.html

"""
from textwrap import dedent
import numpy as np
from scipy.linalg import inv
from ._lss import LinearStateSpace
from ._matrix_eqn import solve_discrete_riccati


class Kalman:
    r"""
    Implements the Kalman filter for the Gaussian state space model

    .. math::

        x_{t+1} = A x_t + C w_{t+1} \\
        y_t = G x_t + H v_t

    Here :math:`x_t` is the hidden state and :math:`y_t` is the measurement.
    The shocks :math:`w_t` and :math:`v_t` are iid standard normals. Below
    we use the notation

    .. math::

        Q := CC'
        R := HH'


    Parameters
    ----------
    ss : instance of LinearStateSpace
        An instance of the quantecon.lss.LinearStateSpace class
    x_hat : scalar(float) or array_like(float), optional(default=None)
        An n x 1 array representing the mean x_hat of the
        prior/predictive density.  Set to zero if not supplied.
    Sigma : scalar(float) or array_like(float), optional(default=None)
        An n x n array representing the covariance matrix Sigma of
        the prior/predictive density.  Must be positive definite.
        Set to the identity if not supplied.

    Attributes
    ----------
    Sigma, x_hat : as above
    Sigma_infinity : array_like or scalar(float)
        The infinite limit of Sigma_t
    K_infinity : array_like or scalar(float)
        The stationary Kalman gain.


    References
    ----------

    https://lectures.quantecon.org/py/kalman.html

    """

    def __init__(self, ss, x_hat=None, Sigma=None):
        self.ss = ss
        self.set_state(x_hat, Sigma)
        self._K_infinity = None
        self._Sigma_infinity = None

    def set_state(self, x_hat, Sigma):
        if Sigma is None:
            self.Sigma = np.identity(self.ss.n)
        else:
            self.Sigma = np.atleast_2d(Sigma)
        if x_hat is None:
            self.x_hat = np.zeros((self.ss.n, 1))
        else:
            self.x_hat = np.atleast_2d(x_hat)
            self.x_hat.shape = self.ss.n, 1

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        m = """\
        Kalman filter:
          - dimension of state space          : {n}
          - dimension of observation equation : {k}
        """
        return dedent(m.format(n=self.ss.n, k=self.ss.k))

    @property
    def Sigma_infinity(self):
        if self._Sigma_infinity is None:
            self.stationary_values()
        return self._Sigma_infinity

    @property
    def K_infinity(self):
        if self._K_infinity is None:
            self.stationary_values()
        return self._K_infinity

    def whitener_lss(self):
        r"""
        This function takes the linear state space system
        that is an input to the Kalman class and it converts
        that system to the time-invariant whitener represenation
        given by

        .. math::

            \tilde{x}_{t+1}^* = \tilde{A} \tilde{x} + \tilde{C} v
            a = \tilde{G} \tilde{x}

        where

        .. math::

            \tilde{x}_t = [x+{t}, \hat{x}_{t}, v_{t}]

        and

        .. math::

            \tilde{A} =
            \begin{bmatrix}
            A  & 0    & 0  \\
            KG & A-KG & KH \\
            0  & 0    & 0 \\
            \end{bmatrix}

        .. math::

            \tilde{C} =
            \begin{bmatrix}
            C & 0 \\
            0 & 0 \\
            0 & I \\
            \end{bmatrix}

        .. math::

            \tilde{G} =
            \begin{bmatrix}
            G & -G & H \\
            \end{bmatrix}

        with :math:`A, C, G, H` coming from the linear state space system
        that defines the Kalman instance

        Returns
        -------
        whitened_lss : LinearStateSpace
            This is the linear state space system that represents
            the whitened system
        """
        K = self.K_infinity

        # Get the matrix sizes
        n, k, m, l = self.ss.n, self.ss.k, self.ss.m, self.ss.l
        A, C, G, H = self.ss.A, self.ss.C, self.ss.G, self.ss.H

        Atil = np.vstack([np.hstack([A, np.zeros((n, n)), np.zeros((n, l))]),
                          np.hstack([np.dot(K, G),
                                     A-np.dot(K, G),
                                     np.dot(K, H)]),
                          np.zeros((l, 2*n + l))])

        Ctil = np.vstack([np.hstack([C, np.zeros((n, l))]),
                          np.zeros((n, m+l)),
                          np.hstack([np.zeros((l, m)), np.eye(l)])])

        Gtil = np.hstack([G, -G, H])

        whitened_lss = LinearStateSpace(Atil, Ctil, Gtil)
        self.whitened_lss = whitened_lss

        return whitened_lss

    def prior_to_filtered(self, y):
        r"""
        Updates the moments (x_hat, Sigma) of the time t prior to the
        time t filtering distribution, using current measurement :math:`y_t`.

        The updates are according to

        .. math::

            \hat{x}^F = \hat{x} + \Sigma G' (G \Sigma G' + R)^{-1}
                (y - G \hat{x})
            \Sigma^F = \Sigma - \Sigma G' (G \Sigma G' + R)^{-1} G
                \Sigma

        Parameters
        ----------
        y : scalar or array_like(float)
            The current measurement

        """
        # === simplify notation === #
        G, H = self.ss.G, self.ss.H
        R = np.dot(H, H.T)

        # === and then update === #
        y = np.atleast_2d(y)
        y.shape = self.ss.k, 1
        E = np.dot(self.Sigma, G.T)
        F = np.dot(np.dot(G, self.Sigma), G.T) + R
        M = np.dot(E, inv(F))
        self.x_hat = self.x_hat + np.dot(M, (y - np.dot(G, self.x_hat)))
        self.Sigma = self.Sigma - np.dot(M, np.dot(G,  self.Sigma))

    def filtered_to_forecast(self):
        """
        Updates the moments of the time t filtering distribution to the
        moments of the predictive distribution, which becomes the time
        t+1 prior

        """
        # === simplify notation === #
        A, C = self.ss.A, self.ss.C
        Q = np.dot(C, C.T)

        # === and then update === #
        self.x_hat = np.dot(A, self.x_hat)
        self.Sigma = np.dot(A, np.dot(self.Sigma, A.T)) + Q

    def update(self, y):
        """
        Updates x_hat and Sigma given k x 1 ndarray y.  The full
        update, from one period to the next

        Parameters
        ----------
        y : np.ndarray
            A k x 1 ndarray y representing the current measurement

        """
        self.prior_to_filtered(y)
        self.filtered_to_forecast()

    def stationary_values(self, method='doubling'):
        r"""
        Computes the limit of :math:`\Sigma_t` as t goes to infinity by
        solving the associated Riccati equation. The outputs are stored in the
        attributes `K_infinity` and `Sigma_infinity`. Computation is via the
        doubling algorithm (default) or a QZ decomposition method (see the
        documentation in `matrix_eqn.solve_discrete_riccati`).

        Parameters
        ----------
        method : str, optional(default="doubling")
            Solution method used in solving the associated Riccati
            equation, str in {'doubling', 'qz'}.

        Returns
        -------
        Sigma_infinity : array_like or scalar(float)
            The infinite limit of :math:`\Sigma_t`
        K_infinity : array_like or scalar(float)
            The stationary Kalman gain.

        """

        # === simplify notation === #
        A, C, G, H = self.ss.A, self.ss.C, self.ss.G, self.ss.H
        Q, R = np.dot(C, C.T), np.dot(H, H.T)

        # === solve Riccati equation, obtain Kalman gain === #
        Sigma_infinity = solve_discrete_riccati(A.T, G.T, Q, R, method=method)
        temp1 = np.dot(np.dot(A, Sigma_infinity), G.T)
        temp2 = inv(np.dot(G, np.dot(Sigma_infinity, G.T)) + R)
        K_infinity = np.dot(temp1, temp2)

        # == record as attributes and return == #
        self._Sigma_infinity, self._K_infinity = Sigma_infinity, K_infinity
        return Sigma_infinity, K_infinity

    def stationary_coefficients(self, j, coeff_type='ma'):
        """
        Wold representation moving average or VAR coefficients for the
        steady state Kalman filter.

        Parameters
        ----------
        j : int
            The lag length
        coeff_type : string, either 'ma' or 'var' (default='ma')
            The type of coefficent sequence to compute.  Either 'ma' for
            moving average or 'var' for VAR.
        """
        # == simplify notation == #
        A, G = self.ss.A, self.ss.G
        K_infinity = self.K_infinity
        # == compute and return coefficients == #
        coeffs = []
        i = 1
        if coeff_type == 'ma':
            coeffs.append(np.identity(self.ss.k))
            P_mat = A
            P = np.identity(self.ss.n)  # Create a copy
        elif coeff_type == 'var':
            coeffs.append(np.dot(G, K_infinity))
            P_mat = A - np.dot(K_infinity, G)
            P = np.copy(P_mat)  # Create a copy
        else:
            raise ValueError("Unknown coefficient type")
        while i <= j:
            coeffs.append(np.dot(np.dot(G, P), K_infinity))
            P = np.dot(P, P_mat)
            i += 1
        return coeffs

    def stationary_innovation_covar(self):
        # == simplify notation == #
        H, G = self.ss.H, self.ss.G
        R = np.dot(H, H.T)
        Sigma_infinity = self.Sigma_infinity

        return np.dot(G, np.dot(Sigma_infinity, G.T)) + R



"""
Tests for markov/random.py

"""
import numpy as np
from numpy.testing import (
    assert_array_equal, assert_raises, assert_array_almost_equal_nulp,
    assert_
)

from quantecon.markov import (
    random_markov_chain, random_stochastic_matrix, random_discrete_dp
)


def test_random_markov_chain_dense():
    sparse = False
    n, k = 5, 3
    mc_dicts = [{'P': random_markov_chain(n, sparse=sparse).P, 'k': n},
                {'P': random_markov_chain(n, k, sparse=sparse).P, 'k': k}]
    for mc_dict in mc_dicts:
        P = mc_dict['P']
        assert_array_equal(P.shape, (n, n))
        assert_array_equal((P > 0).sum(axis=1), [mc_dict['k']]*n)


def test_random_markov_chain_sparse():
    sparse = True
    n, k = 5, 3
    mc_dicts = [{'P': random_markov_chain(n, sparse=sparse).P, 'k': n},
                {'P': random_markov_chain(n, k, sparse=sparse).P, 'k': k}]
    for mc_dict in mc_dicts:
        P = mc_dict['P']
        assert_array_equal(P.shape, (n, n))
        assert_array_equal(P.getnnz(axis=1), [mc_dict['k']]*n)


def test_random_markov_chain_value_error():
    # n <= 0
    assert_raises(ValueError, random_markov_chain, 0)

    # k = 0
    assert_raises(ValueError, random_markov_chain, 2, 0)

    # k > n
    assert_raises(ValueError, random_markov_chain, 2, 3)


def test_random_stochastic_matrix_dense():
    sparse = False
    n, k = 5, 3
    Ps = [random_stochastic_matrix(n, sparse=sparse),
          random_stochastic_matrix(n, k, sparse=sparse)]
    for P in Ps:
        assert_(np.all(P >= 0))
        assert_array_almost_equal_nulp(P.sum(axis=1), np.ones(n))


def test_random_stochastic_matrix_sparse():
    sparse = True
    n, k = 5, 3
    Ps = [random_stochastic_matrix(n, sparse=sparse),
          random_stochastic_matrix(n, k, sparse=sparse)]
    for P in Ps:
        assert_(np.all(P.data >= 0))
        assert_array_almost_equal_nulp(P.sum(axis=1), np.ones(n))


def test_random_stochastic_matrix_dense_vs_sparse():
    n, k = 10, 5
    seed = 1234
    P_dense = random_stochastic_matrix(n, sparse=False, random_state=seed)
    P_sparse = random_stochastic_matrix(n, sparse=True, random_state=seed)
    assert_array_equal(P_dense, P_sparse.toarray())

    P_dense = random_stochastic_matrix(n, k, sparse=False, random_state=seed)
    P_sparse = random_stochastic_matrix(n, k, sparse=True, random_state=seed)
    assert_array_equal(P_dense, P_sparse.toarray())


def test_random_stochastic_matrix_k_1():
    n, k = 3, 1
    P_dense = random_stochastic_matrix(n, k, sparse=False)
    P_sparse = random_stochastic_matrix(n, k, sparse=True)
    assert_array_equal(P_dense[P_dense != 0], np.ones(n))
    assert_array_equal(P_sparse.data, np.ones(n))
    for P in [P_dense, P_sparse]:
        assert_array_almost_equal_nulp(P.sum(axis=1), np.ones(n))


class TestRandomDiscreteDP:
    def setup_method(self):
        self.num_states, self.num_actions = 5, 4
        self.num_sa = self.num_states * self.num_actions
        self.k = 3
        seed = 1234

        self.ddp = \
            random_discrete_dp(self.num_states, self.num_actions, k=self.k,
                               sparse=False, sa_pair=False, random_state=seed)

        labels = ['dense', 'sparse']
        self.ddps_sa = {}
        for label in labels:
            is_sparse = (label == 'sparse')
            self.ddps_sa[label] = \
                random_discrete_dp(self.num_states, self.num_actions, k=self.k,
                                   sparse=is_sparse, sa_pair=True,
                                   random_state=seed)

    def test_shape(self):
        n, m, L = self.num_states, self.num_actions, self.num_sa

        assert_(self.ddp.R.shape == (n, m))
        assert_(self.ddp.Q.shape == (n, m, n))

        for ddp in self.ddps_sa.values():
            assert_(ddp.R.shape == (L,))
            assert_(ddp.Q.shape == (L, n))

    def test_nonzero(self):
        n, m, L, k = self.num_states, self.num_actions, self.num_sa, self.k

        assert_array_equal((self.ddp.Q > 0).sum(axis=-1), np.ones((n, m))*k)
        assert_array_equal((self.ddps_sa['dense'].Q > 0).sum(axis=-1),
                           np.ones(L)*k)
        assert_array_equal(self.ddps_sa['sparse'].Q.getnnz(axis=-1),
                           np.ones(L)*k)

    def test_equal_reward(self):
        assert_array_equal(self.ddp.R.ravel(), self.ddps_sa['dense'].R)
        assert_array_equal(self.ddps_sa['dense'].R, self.ddps_sa['sparse'].R)

    def test_equal_probability(self):
        assert_array_equal(self.ddp.Q.ravel(), self.ddps_sa['dense'].Q.ravel())
        assert_array_equal(self.ddps_sa['dense'].Q,
                           self.ddps_sa['sparse'].Q.toarray())

    def test_equal_beta(self):
        for ddp in self.ddps_sa.values():
            assert_(ddp.beta == self.ddp.beta)



"""
Tests for Numba support utilities

"""
import numpy as np
from numpy.testing import assert_array_equal, assert_
from numba import jit
from quantecon.util.numba import _numba_linalg_solve, comb_jit


@jit(nopython=True)
def _numba_linalg_solve_jitted(a, b):
    return _numba_linalg_solve(a, b)


@jit(nopython=True)
def numba_linalg_solve_orig(a, b):
    return np.linalg.solve(a, b)


class TestNumbaLinalgSolve:
    def setup_method(self):
        self.dtypes = [np.float32, np.float64]
        self.a = np.array([[3, 2, 0], [1, -1, 0], [0, 5, 1]])
        self.b_1dim = np.array([2, 4, -1])
        self.b_2dim = np.array([[2, 3], [4, 1], [-1, 0]])
        self.a_singular = np.array([[0, 1, 2], [3, 4, 5], [3, 5, 7]])

    def test_b_1dim(self):
        for dtype in self.dtypes:
            a = np.asfortranarray(self.a, dtype=dtype)
            b = np.asfortranarray(self.b_1dim, dtype=dtype)
            sol_orig = numba_linalg_solve_orig(a, b)
            r = _numba_linalg_solve_jitted(a, b)
            assert_(r == 0)
            assert_array_equal(b, sol_orig)

    def test_b_2dim(self):
        for dtype in self.dtypes:
            a = np.asfortranarray(self.a, dtype=dtype)
            b = np.asfortranarray(self.b_2dim, dtype=dtype)
            sol_orig = numba_linalg_solve_orig(a, b)
            r = _numba_linalg_solve_jitted(a, b)
            assert_(r == 0)
            assert_array_equal(b, sol_orig)

    def test_singular_a(self):
        for b in [self.b_1dim, self.b_2dim]:
            for dtype in self.dtypes:
                a = np.asfortranarray(self.a_singular, dtype=dtype)
                b = np.asfortranarray(b, dtype=dtype)
                r = _numba_linalg_solve_jitted(a, b)
                assert_(r != 0)


class TestCombJit:
    def setup_method(self):
        self.MAX_INTP = np.iinfo(np.intp).max

    def test_comb(self):
        N, k = 10, 3
        N_choose_k = 120
        assert_(comb_jit(N, k) == N_choose_k)

    def test_comb_zeros(self):
        assert_(comb_jit(2, 3) == 0)
        assert_(comb_jit(-1, 3) == 0)
        assert_(comb_jit(2, -1) == 0)

        assert_(comb_jit(self.MAX_INTP, 2) == 0)

        N = np.intp(self.MAX_INTP**0.5 * 2**0.5) + 1
        assert_(comb_jit(N, 2) == 0)

    def test_max_intp(self):
        assert_(comb_jit(self.MAX_INTP, 0) == 1)
        assert_(comb_jit(self.MAX_INTP, 1) == self.MAX_INTP)



import numpy as np
from numba import njit

@njit
def brent_max(func, a, b, args=(), xtol=1e-5, maxiter=500):
    """
    Uses a jitted version of the maximization routine from SciPy's fminbound.
    The algorithm is identical except that it's been switched to maximization
    rather than minimization, and the tests for convergence have been stripped
    out to allow for jit compilation.

    Note that the input function `func` must be jitted or the call will fail.

    Parameters
    ----------
    func : jitted function
    a : scalar
        Lower bound for search
    b : scalar
        Upper bound for search
    args : tuple, optional
        Extra arguments passed to the objective function.
    maxiter : int, optional
        Maximum number of iterations to perform.
    xtol : float, optional
        Absolute error in solution `xopt` acceptable for convergence.

    Returns
    -------
    xf : float
        The maximizer
    fval : float
        The maximum value attained
    info : tuple
        A tuple of the form (status_flag, num_iter).  Here status_flag
        indicates whether or not the maximum number of function calls was
        attained.  A value of 0 implies that the maximum was not hit.
        The value `num_iter` is the number of function calls.

    Examples
    --------
    >>> @njit
    ... def f(x):
    ...     return -(x + 2.0)**2 + 1.0
    ...
    >>> xf, fval, info = brent_max(f, -2, 2)

    """
    if not np.isfinite(a):
        raise ValueError("a must be finite.")

    if not np.isfinite(b):
        raise ValueError("b must be finite.")

    if not a < b:
        raise ValueError("a must be less than b.")

    maxfun = maxiter
    status_flag = 0

    sqrt_eps = np.sqrt(2.2e-16)
    golden_mean = 0.5 * (3.0 - np.sqrt(5.0))

    fulc = a + golden_mean * (b - a)
    nfc, xf = fulc, fulc
    rat = e = 0.0
    x = xf
    fx = -func(x, *args)
    num = 1

    ffulc = fnfc = fx
    xm = 0.5 * (a + b)
    tol1 = sqrt_eps * np.abs(xf) + xtol / 3.0
    tol2 = 2.0 * tol1

    while (np.abs(xf - xm) > (tol2 - 0.5 * (b - a))):
        golden = 1
        # Check for parabolic fit
        if np.abs(e) > tol1:
            golden = 0
            r = (xf - nfc) * (fx - ffulc)
            q = (xf - fulc) * (fx - fnfc)
            p = (xf - fulc) * q - (xf - nfc) * r
            q = 2.0 * (q - r)
            if q > 0.0:
                p = -p
            q = np.abs(q)
            r = e
            e = rat

            # Check for acceptability of parabola
            if ((np.abs(p) < np.abs(0.5*q*r)) and (p > q*(a - xf)) and
                    (p < q * (b - xf))):
                rat = (p + 0.0) / q
                x = xf + rat

                if ((x - a) < tol2) or ((b - x) < tol2):
                    si = np.sign(xm - xf) + ((xm - xf) == 0)
                    rat = tol1 * si
            else:      # do a golden section step
                golden = 1

        if golden:  # Do a golden-section step
            if xf >= xm:
                e = a - xf
            else:
                e = b - xf
            rat = golden_mean*e

        if rat == 0:
            si = np.sign(rat) + 1
        else:
            si = np.sign(rat)

        x = xf + si * np.maximum(np.abs(rat), tol1)
        fu = -func(x, *args)
        num += 1

        if fu <= fx:
            if x >= xf:
                a = xf
            else:
                b = xf
            fulc, ffulc = nfc, fnfc
            nfc, fnfc = xf, fx
            xf, fx = x, fu
        else:
            if x < xf:
                a = x
            else:
                b = x
            if (fu <= fnfc) or (nfc == xf):
                fulc, ffulc = nfc, fnfc
                nfc, fnfc = x, fu
            elif (fu <= ffulc) or (fulc == xf) or (fulc == nfc):
                fulc, ffulc = x, fu

        xm = 0.5 * (a + b)
        tol1 = sqrt_eps * np.abs(xf) + xtol / 3.0
        tol2 = 2.0 * tol1

        if num >= maxfun:
            status_flag = 1
            break

    fval = -fx
    info = status_flag, num

    return xf, fval, info



r"""
Base class for solving initial value problems (IVPs) of the form:

.. math::

    \frac{dy}{dt} = f(t,y),\ y(t_0) = y_0

using finite difference methods. The `quantecon.ivp` class uses various
integrators from the `scipy.integrate.ode` module to perform the
integration (i.e., solve the ODE) and parametric B-spline interpolation
from `scipy.interpolate` to approximate the value of the solution
between grid points. The `quantecon.ivp` module also provides a method
for computing the residual of the solution which can be used for
assessing the overall accuracy of the approximated solution.

"""
import numpy as np
from scipy import integrate, interpolate


class IVP(integrate.ode):

    r"""
    Creates an instance of the IVP class.

    Parameters
    ----------
    f : callable ``f(t, y, *f_args)``
        Right hand side of the system of equations defining the ODE.
        The independent variable, ``t``, is a ``scalar``; ``y`` is
        an ``ndarray`` of dependent variables with ``y.shape ==
        (n,)``. The function `f` should return a ``scalar``,
        ``ndarray`` or ``list`` (but not a ``tuple``).
    jac : callable ``jac(t, y, *jac_args)``, optional(default=None)
        Jacobian of the right hand side of the system of equations
        defining the ODE.

        .. :math:

            \mathcal{J}_{i,j} = \bigg[\frac{\partial f_i}{\partial y_j}\bigg]

    """

    def __init__(self, f, jac=None):

        super(IVP, self).__init__(f, jac)

    def _integrate_fixed_trajectory(self, h, T, step, relax):
        """Generates a solution trajectory of fixed length."""
        # initialize the solution using initial condition
        solution = np.hstack((self.t, self.y))

        while self.successful():

            self.integrate(self.t + h, step, relax)
            current_step = np.hstack((self.t, self.y))
            solution = np.vstack((solution, current_step))

            if (h > 0) and (self.t >= T):
                break
            elif (h < 0) and (self.t <= T):
                break
            else:
                continue

        return solution

    def _integrate_variable_trajectory(self, h, g, tol, step, relax):
        """Generates a solution trajectory of variable length."""
        # initialize the solution using initial condition
        solution = np.hstack((self.t, self.y))

        while self.successful():

            self.integrate(self.t + h, step, relax)
            current_step = np.hstack((self.t, self.y))
            solution = np.vstack((solution, current_step))

            if g(self.t, self.y, *self.f_params) < tol:
                break
            else:
                continue

        return solution

    def _initialize_integrator(self, t0, y0, integrator, **kwargs):
        """Initializes the integrator prior to integration."""
        # set the initial condition
        self.set_initial_value(y0, t0)

        # select the integrator
        self.set_integrator(integrator, **kwargs)

    def compute_residual(self, traj, ti, k=3, ext=2):
        r"""
        The residual is the difference between the derivative of the B-spline
        approximation of the solution trajectory and the right-hand side of the
        original ODE evaluated along the approximated solution trajectory.

        Parameters
        ----------
        traj : array_like (float)
            Solution trajectory providing the data points for constructing the
            B-spline representation.
        ti : array_like (float)
            Array of values for the independent variable at which to
            interpolate the value of the B-spline.
        k : int, optional(default=3)
            Degree of the desired B-spline. Degree must satisfy
            :math:`1 \le k \le 5`.
        ext : int, optional(default=2)
            Controls the value of returned elements for outside the
            original knot sequence provided by traj. For extrapolation, set
            `ext=0`; `ext=1` returns zero; `ext=2` raises a `ValueError`.

        Returns
        -------
        residual : array (float)
            Difference between the derivative of the B-spline approximation
            of the solution trajectory and the right-hand side of the ODE
            evaluated along the approximated solution trajectory.

        """
        # B-spline approximations of the solution and its derivative
        soln = self.interpolate(traj, ti, k, 0, ext)
        deriv = self.interpolate(traj, ti, k, 1, ext)

        # rhs of ode evaluated along approximate solution
        T = ti.size
        rhs_ode = np.vstack([self.f(ti[i], soln[i, 1:], *self.f_params)
                            for i in range(T)])
        rhs_ode = np.hstack((ti[:, np.newaxis], rhs_ode))

        # should be roughly zero everywhere (if approximation is any good!)
        residual = deriv - rhs_ode

        return residual

    def solve(self, t0, y0, h=1.0, T=None, g=None, tol=None,
              integrator='dopri5', step=False, relax=False, **kwargs):
        r"""
        Solve the IVP by integrating the ODE given some initial condition.

        Parameters
        ----------
        t0 : float
            Initial condition for the independent variable.
        y0 : array_like (float, shape=(n,))
            Initial condition for the dependent variables.
        h : float, optional(default=1.0)
            Step-size for computing the solution. Can be positive or negative
            depending on the desired direction of integration.
        T : int, optional(default=None)
            Terminal value for the independent variable. One of either `T`
            or `g` must be specified.
        g : callable ``g(t, y, f_args)``, optional(default=None)
            Provides a stopping condition for the integration. If specified
            user must also specify a stopping tolerance, `tol`.
        tol : float, optional (default=None)
            Stopping tolerance for the integration. Only required if `g` is
            also specifed.
        integrator : str, optional(default='dopri5')
            Must be one of 'vode', 'lsoda', 'dopri5', or 'dop853'
        step : bool, optional(default=False)
            Allows access to internal steps for those solvers that use adaptive
            step size routines. Currently only 'vode', 'zvode', and 'lsoda'
            support `step=True`.
        relax : bool, optional(default=False)
            Currently only 'vode', 'zvode', and 'lsoda' support `relax=True`.
        **kwargs : dict, optional(default=None)
            Dictionary of integrator specific keyword arguments. See the
            Notes section of the docstring for `scipy.integrate.ode` for a
            complete description of solver specific keyword arguments.

        Returns
        -------
        solution: ndarray (float)
            Simulated solution trajectory.

        """
        self._initialize_integrator(t0, y0, integrator, **kwargs)

        if (g is not None) and (tol is not None):
            soln = self._integrate_variable_trajectory(h, g, tol, step, relax)
        elif T is not None:
            soln = self._integrate_fixed_trajectory(h, T, step, relax)
        else:
            mesg = "Either both 'g' and 'tol', or 'T' must be specified."
            raise ValueError(mesg)

        return soln

    def interpolate(self, traj, ti, k=3, der=0, ext=2):
        r"""
        Parametric B-spline interpolation in N-dimensions.

        Parameters
        ----------
        traj : array_like (float)
            Solution trajectory providing the data points for constructing the
            B-spline representation.
        ti : array_like (float)
            Array of values for the independent variable at which to
            interpolate the value of the B-spline.
        k : int, optional(default=3)
            Degree of the desired B-spline. Degree must satisfy
            :math:`1 \le k \le 5`.
        der : int, optional(default=0)
            The order of derivative of the spline to compute (must be less
            than or equal to `k`).
        ext : int, optional(default=2) Controls the value of returned elements
            for outside the original knot sequence provided by traj. For
            extrapolation, set `ext=0`; `ext=1` returns zero; `ext=2` raises a
            `ValueError`.

        Returns
        -------
        interp_traj: ndarray (float)
            The interpolated trajectory.

        """
        # array of parameter values
        u = traj[:, 0]

        # build list of input arrays
        n = traj.shape[1]
        x = [traj[:, i] for i in range(1, n)]

        # construct the B-spline representation (s=0 forces interpolation!)
        tck, t = interpolate.splprep(x, u=u, k=k, s=0)

        # evaluate the B-spline (returns a list)
        out = interpolate.splev(ti, tck, der, ext)

        # convert to a 2D array
        interp_traj = np.hstack((ti[:, np.newaxis], np.array(out).T))

        return interp_traj



"""
Tests for minmax

"""
import numpy as np
from numpy.testing import assert_, assert_allclose

from quantecon.optimize import minmax
from quantecon.game_theory import NormalFormGame, Player, lemke_howson


class TestMinmax:
    def test_RPS(self):
        A = np.array(
            [[0, 1, -1],
             [-1, 0, 1],
             [1, -1, 0],
             [-1, -1, -1]]
        )
        v_expected = 0
        x_expected = [1/3, 1/3, 1/3, 0]
        y_expected = [1/3, 1/3, 1/3]

        v, x, y = minmax(A)

        assert_allclose(v, v_expected)
        assert_allclose(x, x_expected)
        assert_allclose(y, y_expected)

    def test_random_matrix(self):
        seed = 12345
        rng = np.random.default_rng(seed)
        size = (10, 15)
        A = rng.normal(size=size)
        v, x, y = minmax(A)

        for z in [x, y]:
            assert_((z >= 0).all())
            assert_allclose(z.sum(), 1)

        g = NormalFormGame((Player(A), Player(-A.T)))
        NE = lemke_howson(g)
        assert_allclose(v, NE[0] @ A @ NE[1])
        assert_(g.is_nash((x, y)))



import numpy as np
from numpy.testing import assert_almost_equal, assert_allclose
from numba import njit

from quantecon.optimize import (
    newton, newton_halley, newton_secant, bisect, brentq
)


@njit
def func(x):
    """
    Function for testing on.
    """
    return (x**3 - 1)


@njit
def func_prime(x):
    """
    Derivative for func.
    """
    return (3*x**2)


@njit
def func_prime2(x):
    """
    Second order derivative for func.
    """
    return 6*x


@njit
def func_two(x):
    """
    Harder function for testing on.
    """
    return np.sin(4 * (x - 1/4)) + x + x**20 - 1


@njit
def func_two_prime(x):
    """
    Derivative for func_two.
    """
    return 4*np.cos(4*(x - 1/4)) + 20*x**19 + 1


@njit
def func_two_prime2(x):
    """
    Second order derivative for func_two
    """
    return 380*x**18 - 16*np.sin(4*(x - 1/4))


def test_newton_basic():
    """
    Uses the function f defined above to test the scalar maximization
    routine.
    """
    true_fval = 1.0
    fval = newton(func, 5, func_prime)
    assert_almost_equal(true_fval, fval.root, decimal=4)


def test_newton_basic_two():
    """
    Uses the function f defined above to test the scalar maximization
    routine.
    """
    true_fval = 1.0
    fval = newton(func, 5, func_prime)
    assert_allclose(true_fval, fval.root, rtol=1e-5, atol=0)


def test_newton_hard():
    """
    Harder test for convergence.
    """
    true_fval = 0.408
    fval = newton(func_two, 0.4, func_two_prime)
    assert_allclose(true_fval, fval.root, rtol=1e-5, atol=0.01)


def test_halley_basic():
    """
    Basic test for halley method
    """
    true_fval = 1.0
    fval = newton_halley(func, 5, func_prime, func_prime2)
    assert_almost_equal(true_fval, fval.root, decimal=4)


def test_halley_hard():
    """
    Harder test for halley method
    """
    true_fval = 0.408
    fval = newton_halley(func_two, 0.4, func_two_prime, func_two_prime2)
    assert_allclose(true_fval, fval.root, rtol=1e-5, atol=0.01)


def test_secant_basic():
    """
    Basic test for secant option.
    """
    true_fval = 1.0
    fval = newton_secant(func, 5)
    assert_allclose(true_fval, fval.root, rtol=1e-5, atol=0.001)


def test_secant_hard():
    """
    Harder test for convergence for secant function.
    """
    true_fval = 0.408
    fval = newton_secant(func_two, 0.4)
    assert_allclose(true_fval, fval.root, rtol=1e-5, atol=0.01)


def run_check(method, name):
    a = -1
    b = np.sqrt(3)
    true_fval = 0.408
    r = method(func_two, a, b)
    assert_allclose(true_fval, r.root, atol=0.01, rtol=1e-5,
                    err_msg='method %s' % name)


def test_bisect_basic():
    run_check(bisect, 'bisect')


def test_brentq_basic():
    run_check(brentq, 'brentq')



import numpy as np
import numbers
from .normal_form_game import NormalFormGame
from ..util import check_random_state, rng_integers
from .random import random_pure_actions


class LogitDynamics:
    """
    Class representing the logit-response dynamics model.

    Parameters
    ----------
    data : NormalFormGame or array_like
        The game played in the logit-response dynamics model.

    beta : scalar(float)
        The level of noise in player's decision.

    Attributes
    ----------
    N : scalar(int)
        The number of players in the game.

    players : list(Player)
        The list consisting of all players with the given payoff matrix.

    nums_actions : tuple(int)
        Tuple of the number of actions, one for each player.

    beta : scalar(float)
        See parameters.

    player.logit_choice_cdfs : array_like(float)
        The choice probability of each actions given opponents' actions.

    """
    def __init__(self, data, beta=1.0):
        if isinstance(data, NormalFormGame):
            self.g = data
        else:  # data must be array_like
            self.g = NormalFormGame(data)

        self.N = self.g.N
        self.players = self.g.players
        self.nums_actions = self.g.nums_actions

        self.beta = beta

        for player in self.players:
            payoff_array_rotated = \
                player.payoff_array.transpose(list(range(1, self.N)) + [0])
            payoff_array_rotated -= \
                payoff_array_rotated.max(axis=-1)[..., np.newaxis]
            player.logit_choice_cdfs = \
                np.exp(payoff_array_rotated*self.beta).cumsum(axis=-1)

    def logit_choice_cdfs(self):
        """
        Return the tuple of choice probabilities.

        """
        return tuple(player.logit_choice_cdfs for player in self.players)

    def _play(self, player_ind, actions, random_state):
        i = player_ind

        # Tuple of the actions of opponent players i+1, ..., N, 0, ..., i-1
        opponent_actions = \
            tuple(actions[i+1:]) + tuple(actions[:i])

        cdf = self.players[i].logit_choice_cdfs[opponent_actions]
        random_value = random_state.random()
        next_action = cdf.searchsorted(random_value*cdf[-1], side='right')

        return next_action

    def play(self, init_actions=None, player_ind_seq=None, num_reps=1,
             random_state=None):
        """
        Return a new action profile which is updated `num_reps` times.

        Parameters
        ----------
        init_actions : tuple(int), optional(default=None)
            The action profile in the initial period. If None, selected
            randomly.

        player_ind_seq : list(int), optional(default=None)
            The sequence of player indices. If None, selected randomly.

        num_reps : scalar(int), optional(default=1)
            The number of iterations.

        random_state : int or np.random.RandomState/Generator, optional
            Random number generator used.

        Returns
        -------
        tuple(int)
            The action profile after iterations.

        """
        random_state = check_random_state(random_state)

        if init_actions is None:
            init_actions = random_pure_actions(self.nums_actions, random_state)
        init_actions = list(init_actions)

        if player_ind_seq is None:
            random_state = check_random_state(random_state)
            player_ind_seq = rng_integers(random_state, self.N, size=num_reps)

        if isinstance(player_ind_seq, numbers.Integral):
            player_ind_seq = [player_ind_seq]

        for t, player_ind in enumerate(player_ind_seq):
            random_state = check_random_state(random_state)
            init_actions[player_ind] = self._play(player_ind, init_actions,
                                                  random_state)

        return tuple(init_actions)

    def time_series(self, ts_length, init_actions=None, random_state=None):
        """
        Return the array representing time series of action profiles.

        Parameters
        ----------
        ts_length : scalar(int)
            The number of iterations.

        init_actions : tuple(int), optional(default=None)
            The action profile in the initial period. If None, selected
            randomly.

        random_state : int or np.random.RandomState/Generator, optional
            Random number generator used.

        Returns
        -------
        ndarray(int)
            The array representing the time series of action profiles.

        """
        if init_actions is None:
            random_state = check_random_state(random_state)
            init_actions = random_pure_actions(self.nums_actions, random_state)
        actions = list(init_actions)

        out = np.empty((ts_length, self.N), dtype=int)
        random_state = check_random_state(random_state)
        player_ind_seq = rng_integers(random_state, self.N, size=ts_length)

        for t, player_ind in enumerate(player_ind_seq):
            out[t, :] = actions[:]
            random_state = check_random_state(random_state)
            actions[player_ind] = self._play(player_ind, actions, random_state)

        return out



# flake8: noqa
"""
Initialization of the optimize subpackage
"""
from .linprog_simplex import (
    linprog_simplex, solve_tableau, get_solution, PivOptions
)
from .lcp_lemke import lcp_lemke
from .minmax import minmax
from .scalar_maximization import brent_max
from .nelder_mead import nelder_mead
from .root_finding import newton, newton_halley, newton_secant, bisect, brentq



"""
Tests for timing.py

"""

import time
from numpy.testing import assert_allclose, assert_
from quantecon.util import tic, tac, toc, loop_timer


class TestTicTacToc:
    def setup_method(self):
        self.h = 0.1
        self.digits = 10

    def test_timer(self):

        tic()

        time.sleep(self.h)
        tm1 = tac()

        time.sleep(self.h)
        tm2 = tac()

        time.sleep(self.h)
        tm3 = toc()

        rtol = 2
        atol = 0.05

        for actual, desired in zip([tm1, tm2, tm3],
                                   [self.h, self.h, self.h*3]):
            assert_allclose(actual, desired, atol=atol, rtol=rtol)

    def test_loop(self):

        def test_function_one_arg(n):
            return time.sleep(n)

        def test_function_two_arg(n, a):
            return time.sleep(n)

        test_one_arg = \
            loop_timer(5, test_function_one_arg, self.h, digits=10)
        test_two_arg = \
            loop_timer(5, test_function_two_arg, [self.h, 1], digits=10)

        rtol = 2
        atol = 0.05

        for tm in test_one_arg:
            assert_allclose(tm, self.h, atol=atol, rtol=rtol)
        for tm in test_two_arg:
            assert_allclose(tm, self.h, atol=atol, rtol=rtol)

        for (average_time, average_of_best) in [test_one_arg, test_two_arg]:
            assert_(average_time >= average_of_best)
